(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: check.ml,v 1.70 2008/02/12 22:13:02 blanqui Exp $ *)

(** {3 Builds various htables for generator and types.} *)

open Parsetree
open Relation
open Longident
;;

(** Possible errors. *)

type error =
   | Unknown_generator of Longident.t
   | No_neutral_element of Longident.t
   | Public_type_with_relations of string
   | Cannot_find_module of Longident.t
   | Not_supported of string
   | Relation_already_declared of Location.t
   | Incompatible_types of Longident.t * Longident.t
   | Invalid_pattern
   | Pattern_not_headed_by_correct_generator of Longident.t
   | Illegal_relation of Longident.t
;;

exception Error of Location.t * error
;;

let error loc err = raise (Error (loc, err))
;;

let unknown_gen g = error g.pgen_loc (Unknown_generator g.pgen_desc)
;;

let relation_already_declared r1 r2 =
  error r1.prel_loc (Relation_already_declared r2.prel_loc)
;;

let no_neutral_element g = error g.pgen_loc (No_neutral_element g.pgen_desc)
;;

(** Error messages. *)

open Format
;;

let report_error ppf = function
  | Unknown_generator li ->
      fprintf ppf "Unknown generator %a." Pr_otype.longident li
  | No_neutral_element li ->
      fprintf ppf "%a has an inverse but no neutral element."
        Pr_otype.longident li
  | Public_type_with_relations s ->
      fprintf ppf "%s is a public type with relations." s
  | Cannot_find_module li ->
      fprintf ppf "Cannot find module %a in search path."
        Pr_otype.longident li
  | Not_supported s ->
      fprintf ppf "%s not supported." s
  | Relation_already_declared loc ->
      fprintf ppf "Relation already declared at %a." Location.print loc
  | Incompatible_types (li1, li2) ->
      fprintf ppf
	"Generators %a and %a have incompatible types."
	Pr_otype.longident li1 Pr_otype.longident li2
  | Invalid_pattern ->
      fprintf ppf "Invalid pattern."
  | Pattern_not_headed_by_correct_generator li ->
      fprintf ppf "Pattern not headed by %a." Pr_otype.longident li
  | Illegal_relation li ->
      fprintf ppf "Relation incompatible with the type of %a."
	Pr_otype.longident li
;;

(** The locality flag for a generator: is it defined in the local module or
    globally ? *)

type locality =
   | Local
   | Global of string
;;

(** Generator information. *)

type generator_info = {
  gi_type_name : string;
  gi_locality : locality;
  gi_arg_types : core_type list;
  gi_relations : relations;
  gi_location : Location.t;
  gi_priv_flag : Asttypes.private_flag;
}
;;

let is_local_info = function
  | { gi_locality = Local } -> true
  | _ -> false
;;

let is_private_info = function
  | { gi_priv_flag = Asttypes.Private } -> true
  | _ -> false
;;

let is_varyadic_info = function
  | { gi_arg_types = [ct] } -> Otype.is_list_type ct
  | _ -> false
;;

let arity_info gi = List.length gi.gi_arg_types
;;

let is_truly_unary_info  = function
  | { gi_arg_types = [ct] } -> not (Otype.is_list_type ct)
  | _ -> false
;;

let find tbl g =
  try Hashtbl.find tbl g.pgen_desc
  with Not_found -> unknown_gen g
;;

(** Generator table. *)

let gname_tbl = Hashtbl.create 97
;;

let find_gident = Hashtbl.find gname_tbl
;;

let find_gname s = find_gident (Lident s)
;;

(** Type table. *)

let tname_tbl = Hashtbl.create 19
;;

let find_tident = Hashtbl.find tname_tbl
;;

let find_tname s = find_tident (Lident s)
;;

(** Various access/test functions. *)

let is_varyadic_gen g = is_varyadic_info (find gname_tbl g)
;;

let is_local_gen g = is_local_info (find gname_tbl g)
;;

let is_private_gen g = is_private_info (find gname_tbl g)
;;

let arity_gen g = arity_info (find gname_tbl g)
;;

let is_truly_unary_gen g = is_truly_unary_info (find gname_tbl g)
;;

let neutral_element_gen g =
  let info = find gname_tbl g in
    match neutral info.gi_relations with
      | Some (_sd, e) -> e
      | _ -> no_neutral_element g
;;

(** Check if a type declaration is admissible. *)

(* here, we should check basic things like:
- a relation is declared only once
- there is no two commutative's
the more complex checks should be done in the completion procedure
*)

let is_possible_for_truly_unary rel =
  match rel.prel_desc with
  | Division_by_Absorbent _
  | Associative _
  | Commutative _
  | Neutral _
  | Absorbing _
  | Inverse _
    -> false
  | Involutive
  | Distributive _
  | Nilpotent _
  | Idempotent _
  | Structure_item _
  | Status _
  | Precedence _
  | Rewrite _
    -> true
;;

let is_only_for_truly_unary rel =
  match rel.prel_desc with
  | Division_by_Absorbent _
  | Associative _
  | Commutative _
  | Neutral _
  | Absorbing _
  | Inverse _
  | Distributive _
  | Nilpotent _
  | Idempotent _
  | Structure_item _
  | Status _
  | Precedence _
  | Rewrite _
    -> false
  | Involutive
    -> true
;;

let is_possible_for_nary rel =
  match rel.prel_desc with
  | Division_by_Absorbent _
  | Associative _
  | Commutative _
  | Neutral _
  | Absorbing _
  | Inverse _
  | Distributive _
  | Nilpotent _
  | Idempotent _
  | Involutive
    -> false
  | Structure_item _
  | Precedence _
  | Rewrite _
  | Status _
    -> true
;;

let rec insert r1 = function
(* 2nd argument is assumed to be sorted wrt compare_relation *)
  | [] -> [r1]
  | (r2 :: rs) as l ->
      begin
        match compare_relation precedence r1 r2 with
          | n when n > 0 -> r2 :: insert r1 rs
          | 0 -> relation_already_declared r1 r2
          | _ -> r1 :: l
      end
;;

let have_compatible_types d e =
  is_varyadic_gen e or (not (is_varyadic_gen d) && arity_gen e = arity_gen d)
;;

let check_relations gname =
  let gid = Lident gname
  and gi = find_gname gname in
  let n = arity_info gi
  and is_truly_unary = is_truly_unary_info gi in
  let rec aux (prec, ass, com, builtins) = function
    | [] -> ()
    | r :: rs ->
	begin
	  let error = error r.prel_loc in
	  if (is_truly_unary && not (is_possible_for_truly_unary r))
	    or (not is_truly_unary && is_only_for_truly_unary r)
	    or (n <> 1 && n <> 2 && not (is_possible_for_nary r)) then
	    error (Illegal_relation gid);
	  match r.prel_desc, prec, ass, com with
	    | Precedence _, None, _, _ -> aux (Some r, ass, com, builtins) rs
	    | Precedence _, Some p, _, _ -> relation_already_declared p r
           (* | Status _, None, _, _ -> aux (Some r, ass, com, builtins) rs
	    | Status _, Some p, _, _ -> relation_already_declared p r
           *)
	    | Associative _, _, None, _ -> aux (prec, Some r, com, builtins) rs
	    | Associative _, _, Some a, _ -> relation_already_declared a r
	    | Commutative _, _, _, None -> aux (prec, ass, Some r, builtins) rs
	    | Commutative _, _, _, Some c -> relation_already_declared c r
	    | Rewrite ({ppat_desc = Ppat_construct (li, _, _)}, _), _, _, _
		when gid <> li ->
		error (Pattern_not_headed_by_correct_generator gid)
	    | Rewrite ({ppat_desc = Ppat_construct _}, _), _, _, _
	    | Structure_item _, _, _, _ -> aux (prec, ass, com, builtins) rs
	    | Rewrite _, _, _, _ -> error Invalid_pattern
	    | Distributive (_s, d, Some e, _b), _, _, _
		when not (have_compatible_types d e) ->
		error (Incompatible_types (d.pgen_desc, e.pgen_desc))
	    | _ -> aux (prec, ass, com, insert r builtins) rs
	end
  in fun rels ->
    match rels.prels_desc with
      | Prels_none
      | Prels_commented _ -> ()
      | Prels_begend rs -> aux (None, None, None, []) rs
;;

let check_cons_decl (gname, _cts, rels, _loc) = check_relations gname rels
;;

let check_type_decl (_tname, td) =
  match td.ptype_kind with
    | Ptype_variant cds -> List.iter check_cons_decl cds
    | Ptype_record _
    | Ptype_abstract _ -> ()
;;

let check_type_decls = List.iter check_type_decl
;;

(** Add type declarations in the tables. *)

let add tbl local_flag (gname, x) =
  Hashtbl.add tbl (Lident gname) x;
  match local_flag with
    | Local -> ()
    | Global s -> Hashtbl.add tbl (Ldot (Lident s, gname)) x
;;

let fill_tname_tbl = add tname_tbl
;;

let fill_gname_tbl local_flag (tname, td) =
  match td.ptype_kind with
    | Ptype_variant cds ->
      let add_cd (gname, cts, rels, loc) =
	let info =
          { gi_type_name = tname;
            gi_locality = local_flag;
            gi_arg_types = cts;
            gi_relations = rels;
            gi_location = loc;
	    gi_priv_flag = td.ptype_private } in
          add gname_tbl local_flag (gname, info)
      in List.iter add_cd cds
   | Ptype_record _
   | Ptype_abstract _ -> ()
;;

(* first add all data in the tables, then check the data
since some checks need the table to be filled (eg. neutral element). *)

let add_type_decls local_flag ntds =
  List.iter
    (fun ntd ->
       fill_tname_tbl local_flag ntd;
       fill_gname_tbl local_flag ntd)
    ntds;
  check_type_decls ntds
;;
