(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: main.ml,v 1.92.2.2 2008/02/13 16:00:27 bonichon Exp $ *)

(** {3 Driving the mocac compiler and its options.} *)

open Parsetree
open Genr_base
open Genr
open Genr_sharing
open Genr_mli
open Printf
open File
open Configuration
;;

exception Usage_error of string
;;

let usage s = raise (Usage_error s)
;;

(* The table of locally open modules. *)

module Str = struct
  type t = string
  let compare = Pervasives.compare
end
;;

module StrSet = Set.Make (Str);;

let add_local_module, is_local_module =
  let module_set = ref StrSet.empty in
  let add_local s = module_set := StrSet.add s !module_set
  and is_local s = StrSet.mem s !module_set in
  add_local, is_local
;;

(* Do we generate an interface ? *)
let get_genr_interface, set_genr_interface =
  let genr_interface = ref false in
  (fun () -> !genr_interface),
  (fun () -> genr_interface := true)
;;

(* Do we generate an implementation ? *)
let get_genr_implementation, set_genr_implementation =
  let genr_implementation = ref false in
  (fun () -> !genr_implementation),
  (fun () -> genr_implementation := true)
;;

(* Do we generate a test file ? *)
let get_genr_test, set_genr_test =
  let genr_test = ref false in
  (fun () -> !genr_test),
  (fun () -> genr_test := true)
;;

(* Number of tests generated per equation. *)
let get_num_tests, set_num_tests =
  let num_tests = ref 5 in
  (fun () -> !num_tests),
  (fun n -> num_tests := n)

(* Maximal constructor depth of each generated value *)
let get_val_depth, set_val_depth =
  let val_depth = ref 3 in
  (fun () -> !val_depth),
  (fun n -> val_depth := n)

(* Do we use completion ? *)
let get_kb, set_kb =
  let kb = ref false in
  (fun () -> !kb),
  (fun () -> kb := true)
;;

(* Maximal number of completion steps. *)
let get_kb_limit, set_kb_limit =
  let limit = ref 5 in
  (fun () -> !limit),
  (fun n -> limit := n)
;;

(* The output file name for the test file. *)
let get_testing_file_name, set_testing_file_name =
  let testing_file_name = ref "-" in
  (fun () -> !testing_file_name),
  (fun s ->
     if !testing_file_name = "-" then testing_file_name := s
     else usage "test file name already set.")
;;

(* The output file name for the implementation of the generated module. *)
let get_output_module_implementation_name,
    set_output_module_implementation_name =
  let output_module_implementation_name = ref "-" in
  (fun () -> !output_module_implementation_name),
  (fun fname ->
     if !output_module_implementation_name = "-"
     then output_module_implementation_name := fname
     else usage "output implementation module name already set.")
;;

(* The output file name for the interface of the generated module. *)
let get_output_module_interface_name, set_output_module_interface_name =
  let output_module_interface_name = ref "-" in
  (fun () -> !output_module_interface_name),
  (fun fname ->
     if !output_module_interface_name = "-"
     then output_module_interface_name := fname
     else usage "output interface module name already set.")
;;

(* The output file name for the generated module. *)
let get_output_module_name, set_output_module_name =
  let output_module_name = ref "-" in
  (fun () -> !output_module_name),
  (fun fname ->
     if !output_module_name = "-" then
       begin
         output_module_name := fname;
         set_output_module_interface_name (!output_module_name ^ ".mli");
         set_output_module_implementation_name (!output_module_name ^ ".ml");
         set_testing_file_name (!output_module_name ^ "_test.ml")
       end
     else usage "output module name already set.")
;;

(* The mlm source file. *)
let get_input_file_name, set_input_file_name =
  let input_file_name = ref "-" in
  (fun () -> !input_file_name),
  (fun fname ->
     if !input_file_name = "-" then input_file_name := fname
     else usage "input file name already set.")
;;

let set_implementation_from_file_name fname =
  set_genr_implementation ();
  set_input_file_name fname
;;

let set_output_file_name fname =
  if get_output_module_interface_name () = "-" then
    set_output_module_interface_name
      (change_mlm_extension "" fname ^ ".mli");
  if get_output_module_implementation_name () = "-" then
    set_output_module_implementation_name
      (change_mlm_extension "" fname ^ ".ml");
  if get_testing_file_name () = "-" then
    set_testing_file_name
      (change_mlm_extension "" fname ^ "_test.ml")
;;

let set_completion_order =
    (fun oname ->
     match oname with
      | "lex" | "lpo" ->  Eqnrel.set_order Eqnrel.LPO
      | "mul" | "mpo" -> Eqnrel.set_order Eqnrel.MPO
      | "rpo" -> Eqnrel.set_order Eqnrel.RPO
      | _ -> usage "Unknown ordering"
     )
;;


(* The functions that checks the admissibility of a .mlm file
   and generates the corresponding interface and implementation. *)

let check_signature_item s si =
  match si.psig_desc with
  | Psig_type ntds -> Check.add_type_decls (Check.Global s) ntds
  | _ -> ()
;;

let check_structure_item s si =
  match si.pstr_desc with
  | Pstr_type ntds -> Check.add_type_decls (Check.Global s) ntds
  | _ -> ()
;;

let check_and_genr_ml sharing ob si =
  let print () = bprintf ob "%a;;\n\n" Pr_otype.verbatim_structure_item si in
  match si.pstr_desc with
  | Pstr_type ntds ->
    Check.add_type_decls Check.Local ntds;
    if sharing
    then Genr_sharing.print_sharing ob ntds
    else Genr_mli.genr_public_type_declarations sharing ob ntds;
    if get_kb () then
      begin
        if Configuration.get_verbose () then
          print_string "Completing equations ...\n";
        Eqnrel.completion (get_kb_limit ()) ntds
      end;
    let structure_items = Genr.genr_functions ntds in
    Pr_code.structure ob structure_items
  | Pstr_module (s, _) ->
      print ();
      add_local_module s
  | Pstr_recmodule l ->
      print ();
      List.iter (fun (s, _, _) -> add_local_module s) l
  | Pstr_open (Longident.Lident s as li) ->
    begin
      print ();
      if not (is_local_module s) then
      try
        let fn = search_mli_file_for_module s in
        let sis = parse_mli_file fn in
        List.iter (check_signature_item s) sis with
      | Not_found ->
        try
          let fn = search_ml_file_for_module s in
          let sis = parse_ml_file fn in
          List.iter (check_structure_item s) sis with
        | Not_found ->
            Check.error si.pstr_loc (Check.Cannot_find_module li)
    end
  | _ ->
    print ()
;;

let check_and_genr_mli sharing ob si =
  match si.pstr_desc with
  | Pstr_type ntds -> genr_mli sharing ob ntds
  | Pstr_module (s, _) -> add_local_module s
  | Pstr_recmodule l -> List.iter (fun (s, _, _) -> add_local_module s) l
  | Pstr_open (Longident.Lident s as lid) when not (is_local_module s) ->
    bprintf ob "open %a;;\n" Pr_otype.blongident lid
  | _ -> ()
;;

let genr_test ob si =
  let module_name = module_of_mli_file (get_output_module_interface_name ()) in
  match si.pstr_desc with
  | Pstr_type ntds ->
      Genr_testing.genr_and_print_test ob
        module_name true (get_num_tests ()) (get_val_depth ()) ntds
  | Pstr_open (Longident.Lident s as lid) when not (is_local_module s) ->
    bprintf ob "open %a;;\n" Pr_otype.blongident lid
  | _ -> ()
;;

(* Parsing of the current mlm file and generation of the corresponding
   interface and implementation files. *)

let ob = Buffer.create 10000
;;

let genr_file print_sharing_preamble genr sis oname =
  let oc = File.careful_open_out oname in
  let sharing = get_sharable_target () in
  if sharing then print_sharing_preamble ob;
  List.iter (genr sharing ob) sis;
  Buffer.output_buffer oc ob;
  Buffer.reset ob;
  File.careful_close_out oc;
;;

let genr_file_mli sis oname =
  if Configuration.get_verbose () then
    print_string "Generating interface ...\n";
  genr_file
    Genr_sharing.print_sharing_mli_preamble
    check_and_genr_mli
    sis oname
;;

let genr_file_ml sis oname =
  if Configuration.get_verbose () then
    print_string "Generating implementation ...\n";
  genr_file
    Genr_sharing.print_sharing_ml_preamble
    check_and_genr_ml
    sis oname
;;

let genr_test_file sis oname =
  let oc = File.careful_open_out oname in
  List.iter (genr_test ob) sis;
  Buffer.output_buffer oc ob;
  Buffer.reset ob;
  File.careful_close_out oc;
;;

let genr fname =
  try

    if Filename.check_suffix fname ".mlms" then
      begin
        if Configuration.get_verbose () then
          print_string ".mlms input file : using sharing ...\n";
        set_sharable_target ();
      end;

    let sis = parse_ml_file fname in

    begin
      match get_genr_interface (), get_genr_implementation () with
        | true, true
        | false, false ->
            genr_file_mli sis (get_output_module_interface_name ());
            genr_file_ml sis (get_output_module_implementation_name ())
        | true, false ->
            genr_file_mli sis (get_output_module_interface_name ())
        | false, true ->
            genr_file_ml sis (get_output_module_implementation_name ())
    end;

    if get_genr_test () then
      begin
        if Configuration.get_verbose () then
          print_string "Generating test file ...\n";
        genr_test_file sis (get_testing_file_name ())
      end

  with

    | Syntaxerr.Error e ->
        Format.fprintf Format.err_formatter "%a@." Syntaxerr.report_error e;
        exit 2

    | Check.Error (loc, e) ->
        Location.print Format.err_formatter loc;
        Format.fprintf Format.err_formatter
          "Checking error: %a@." Check.report_error e;
        exit 2
;;

(* Parsing the command line. *)

let umsg = "Usage: mocac [options] <.mlm[s] file>"
;;

let print_moca_version v =
  prerr_endline
   (sprintf
      "The Moca relational types compiler, version %s" v);
  exit 0
;;

let print_moca_short_version () =
  print_moca_version
    (sprintf "%.2f" Configuration.moca_version_number)
;;

let print_moca_full_version () =
  print_moca_version Configuration.moca_full_version
;;

let set_moca_to_verbose () = Configuration.set_verbose true
;;

let rec argspec =  [
  ("-c", Arg.String set_implementation_from_file_name,
   " generate a module for the input file argument");
  ("-help", Arg.Unit print_usage,
   " print this option list and exit");
  ("--help", Arg.Unit print_usage,
   " print this option list and exit");
  ("-i", Arg.Unit set_genr_interface,
   " generate a module interface for the input file argument");
  ("-I", Arg.String add_dir_to_search_path,
   "<dir>   add a directory to the search path");
  ("-kb", Arg.Unit set_kb,
   " set completion on (experimental option under development)");
  ("-limit", Arg.Int set_kb_limit,
   "<n>   set the upper bound to completion steps");
  ("-ntests", Arg.Int set_num_tests,
   "<n>   set the number of generated tests per equation \
  \n      (use with -test)");
  ("-o", Arg.String set_output_module_name,
   "<file>   output the module with name the file name argument");
  ("-oml", Arg.String set_output_module_implementation_name,
   "<file>   output the module implementation in the file argument");
  ("-omli", Arg.String set_output_module_interface_name,
   "<file>   output the module interface in the file argument");
  ("-order", Arg.String set_completion_order,
   "<name>   set the order for completion (lpo, mpo and rpo are supported)");
  ("-otest", Arg.String set_testing_file_name,
   "<file>   output tests in the file argument");
  ("--sharing", Arg.Unit set_sharable_target,
   " generate maximally sharing data structures");
  ("-test", Arg.Unit set_genr_test,
   " generate a file for testing the code generated by Moca");
  ("-urorder", Arg.Unit Complete.set_user_rel_order,
   " use the relation order provided in the source file");
  ("-valdepth", Arg.Int set_val_depth,
   "<n>   set the maximal constructor depth of the values \
  \n      generated for testing (to be used with -test)");
  ("--verbose", Arg.Unit set_moca_to_verbose,
   " be verbose during generation");
  ("-v", Arg.Unit print_moca_short_version,
   " print version string and exit");
  ("--version", Arg.Unit print_moca_full_version,
   " print full version string and exit");
]

and print_usage () =
  Arg.usage (Arg.align argspec) umsg;
  exit 0;
;;

(* The main procedure *)

let main () =

  try
    Arg.parse (Arg.align argspec) set_input_file_name umsg;
    let fname = get_input_file_name () in
    set_output_file_name fname;
    genr fname;
    exit 0 with
  | Usage_error s ->
      Format.fprintf Format.err_formatter "Usage error: %s@." s;
      exit 2
  | Unknown_file_extension s ->
      Format.fprintf Format.err_formatter
        "Usage error: file ``%s'' has no known extension.@." s;
      exit 2
;;

(* Run the main procedure. *)

main ();;
